#!/usr/bin/perl -w
#
# file    : index.pl
# changed : 2005-06-05
# author  : Bas Couwenberg <sebastic@xs4all.nl>
#

use strict qw(subs vars);
use Cwd;
use Benchmark ':hireswallclock';
use Fcntl ':flock'; # import LOCK_* constants
use File::Basename;
use File::Copy;
use Date::Calc qw(:all);
use Date::Manip;
use Image::Magick;
use XML::Simple;
use Data::Dumper;
use URI::Escape;

# for Basic Authentication
use Apache::Htpasswd;
use MIME::Base64;

my $time_start = new Benchmark;

my $apache2 = 0; 	# assume apache 1.3 by default
#   $apache2 = 1;	# uncomment to explicitly set apache2

# detect apache version of ServerSignature in environment
if($ENV{SERVER_SOFTWARE} =~ /^Apache\/2/i) {
        $apache2 = 1;
}
elsif($ENV{SERVER_SOFTWARE} =~ /^Apache\/1/i) {
	$apache2 = 0;
}

# Apache Request objects
my ($r, $c, $s);
if($apache2 != 1) {
	# Running on Apache 1.3 with mod_perl probably
        
	require Apache::Request;
	import  Apache::Request;
	
	$r = Apache->request;
}
else {
	# Running on Apache 2.x with mod_perl2 probabl
	
	require Apache2::Request;
	require Apache2::Upload;
	require Apache2::compat;
	
	import Apache2::Request;
	import Apache2::Upload;
	import Apache2::compat;
	
	$r = Apache2::RequestUtil->request;
}
$c = $r->connection;
$s = $r->server;

# Initialize global variables
my %args = &get_args($r, $c, $s, $apache2);

# Apache2 doesn't use the directory in which this file is located as
# its Current Working Directory.
# Extract directory from the Filename and change directory to it.
(my $cwd = $args{Filename}) =~ s/^(.*\/).*?$/$1/;
Cwd::chdir $cwd;

   $args{'version'} 	   	        = '1.2.2';

   $args{'www-data-dir'}		= $cwd;
   $args{'rel-www-data-dir'}            = ''; # ./
   $args{'stanford-data-dir'}	 	= 'stanford-data/';
   $args{'project-data-dir'}	 	= 'project-data/';
   $args{'client-data-dir'}	 	= 'client-data/';
   $args{'xml-data-dir'}	 	= 'xml-data/';
   $args{'qd-data-dir'}	       		= 'qd-data/';
   $args{'img-dir'}			= 'img/';
   $args{'bin-dir'}			= 'bin/';
   $args{'settings-dir'}		= 'settings/';
   $args{'queue-graphs-dir'}		= 'queue-graphs/';
   
   $args{'htaccess-file'}               = $cwd.'.htaccess';
   $args{'htpasswd-file'}               = $cwd.'.htpasswd';
   
   $args{'settings-htaccess-file'}      = $cwd.$args{'settings-dir'}.'.htaccess';
   $args{'settings-htpasswd-file'}      = $cwd.$args{'settings-dir'}.'.htpasswd';
   
   $args{'client-data-htaccess-file'}   = $cwd.$args{'client-data-dir'}.'.htaccess';
   $args{'client-data-htpasswd-file'}   = $cwd.$args{'client-data-dir'}.'.htpasswd';
  
# edit the paths to the directories, 
# if this is the index.pl in the settings/ dir
if($cwd =~ /^(.*\/)settings\/$/) {
   my $wwwroot = $1;
   
   $args{'www-data-dir'}		= $wwwroot;
   
   $args{'htaccess-file'}               = $wwwroot.'.htaccess';
   $args{'htpasswd-file'}               = $wwwroot.'.htpasswd';
   
   $args{'settings-htaccess-file'}      = $cwd.'.htaccess';
   $args{'settings-htpasswd-file'}      = $cwd.'.htpasswd';
   
   $args{'client-data-htaccess-file'}   = $wwwroot.$args{'client-data-dir'}.'.htaccess';
   $args{'client-data-htpasswd-file'}   = $wwwroot.$args{'client-data-dir'}.'.htpasswd';

   $args{'www-stanford-data-dir'}	= $args{'stanford-data-dir'};
   $args{'www-project-data-dir'}	= $args{'project-data-dir'};
   $args{'www-client-data-dir'}	 	= $args{'client-data-dir'};
   $args{'www-xml-data-dir'}	 	= $args{'xml-data-dir'};
   $args{'www-qd-data-dir'}		= $args{'qd-data-dir'};
   $args{'www-img-dir'}			= $args{'img-dir'};
   $args{'www-bin-dir'}			= $args{'bin-dir'};
   $args{'www-queue-graphs-dir'}        = $args{'queue-graphs-dir'};
   $args{'www-settings-dir'}		= $args{'settings-dir'};
   
   $args{'rel-www-data-dir'}            = '../';
   $args{'stanford-data-dir'}	 	= $args{'rel-www-data-dir'}.$args{'stanford-data-dir'};
   $args{'project-data-dir'}	 	= $args{'rel-www-data-dir'}.$args{'project-data-dir'};
   $args{'client-data-dir'}	 	= $args{'rel-www-data-dir'}.$args{'client-data-dir'};
   $args{'xml-data-dir'}	 	= $args{'rel-www-data-dir'}.$args{'xml-data-dir'};
   $args{'qd-data-dir'}			= $args{'rel-www-data-dir'}.$args{'qd-data-dir'};
   $args{'img-dir'}			= $args{'rel-www-data-dir'}.$args{'img-dir'};
   $args{'bin-dir'}			= $args{'rel-www-data-dir'}.$args{'bin-dir'};
   $args{'www-queue-graphs-dir'}        = $args{'rel-www-data-dir'}.$args{'queue-graphs-dir'};
   $args{'settings-dir'}		= ''; # ./
}
   
   $args{'client-list-file'}            = $args{'xml-data-dir'}.'client-list.xml';
   $args{'project-list-file'}	 	= $args{'xml-data-dir'}.'project-list.xml';
   $args{'team-list-file'}              = $args{'xml-data-dir'}.'team-list.xml';
   $args{'username-list-file'}	 	= $args{'xml-data-dir'}.'username-list.xml';
   $args{'known-projects-file'}         = $args{'xml-data-dir'}.'known-projects.xml';
   $args{'client-projects-file'}        = $args{'xml-data-dir'}.'client-projects.xml';
   $args{'project-images-file'}         = $args{'xml-data-dir'}.'project-images.xml';
   $args{'xml-project-summary-file'}    = $args{'xml-data-dir'}.'project-summary.xml';
   $args{'stanford-files-file'}         = $args{'xml-data-dir'}.'stanford-files.xml';
   $args{'teamstats-list-file'}         = $args{'xml-data-dir'}.'teamstats-list.xml';
   $args{'qd-files-file'}               = $args{'xml-data-dir'}.'qd-files.xml';
   $args{'fci-summary-file'}            = $args{'xml-data-dir'}.'fci-summary.xml';
   $args{'jmol-missing-projects-file'}  = $args{'xml-data-dir'}.'jmol-missing-projects.xml';
   $args{'jmol-projects-file'}          = $args{'xml-data-dir'}.'jmol-projects.xml';
   $args{'queue-graphs-file'}           = $args{'xml-data-dir'}.'queue-graphs.xml';

   $args{'user-summary-file'}	 	= $args{'stanford-data-dir'}.'daily_user_summary.txt';
   $args{'team-summary-file'}	 	= $args{'stanford-data-dir'}.'daily_team_summary.txt';
   $args{'project-summary-file'} 	= $args{'stanford-data-dir'}.'psummary.html';
   
   $args{'user-summary-url'}            = 'http://fah-web.stanford.edu/daily_user_summary.txt';
   $args{'team-summary-url'}            = 'http://fah-web.stanford.edu/daily_team_summary.txt';
   $args{'project-summary-url'}         = 'http://fah-web.stanford.edu/psummary.html';
   
   $args{'qd-bin'}                      = $args{'qd-data-dir'}.'qd';
   $args{'qd-windows-bin'}              = $args{'qd-data-dir'}.'qd.exe';
   $args{'qdinfo-file'}                 = $args{'qd-data-dir'}.'qdinfo.dat';
   
   $args{'qd-bin-update'}               = $args{'qd-data-dir'}.'qd.lastupdate';
   $args{'qd-windows-bin-update'}       = $args{'qd-data-dir'}.'qd.exe.lastupdate';
   $args{'qdinfo-file-update'}          = $args{'qd-data-dir'}.'qdinfo.dat.lastupdate';
   
   #$args{'qd-base-url'}		= 'http://www.boston.quik.com/rph/';
   #$args{'qd-url'}			= 'http://www.boston.quik.com/rph/fah.html#qd';
   #$args{'qdinfo-url'}			= 'http://www.boston.quik.com/rph/fah.html#qdinfo';
   
   $args{'qd-base-url'}			= 'http://linuxminded.xs4all.nl/mirror/www.boston.quik.com/rph/';
   $args{'qd-url'}			= $args{'qd-base-url'}.'fah.html#qd';
   $args{'qdinfo-url'}			= $args{'qd-base-url'}.'fah.html#qdinfo';
   $args{'xyz2pdb-url'}                 = $args{'qd-base-url'}.'fah.html#xyz2pdb';

   $args{'qd-bin-linux-url'}            = 'http://linuxminded.xs4all.nl/software/qd-tools/binaries/linux/qd';
   $args{'qd-bin-win32-url'}            = 'http://linuxminded.xs4all.nl/software/qd-tools/binaries/win32/qd.exe';
   $args{'qdinfo.dat-url'}              = 'http://linuxminded.xs4all.nl/software/qd-tools/source/qdinfo/qdinfo.dat';
   
   $args{'rasmol-bin'}			= $args{'bin-dir'}.'rasmol_32BIT';
   $args{'xyz2pdb-bin'}			= $args{'bin-dir'}.'xyz2pdb';

   $args{'htpasswd-bin'}                = '/usr/bin/htpasswd';
   $args{'convert-bin'}                 = '/usr/bin/convert';

   $args{'fci-update-stanford-files-bin'} = '/usr/bin/fci-update-stanford-files.pl';
   $args{'fci-update-qd-files-bin'}       = '/usr/bin/fci-update-qd-files.pl';
   $args{'fci-update-xml-files-bin'}      = '/usr/bin/fci-update-xml-files.pl';
   $args{'fci-update-project-images-bin'} = '/usr/bin/fci-update-project-images.pl';
   $args{'fci-jmol-missing-projects-bin'} = '/usr/bin/fci-jmol-missing-projects.pl';
   $args{'fci-generate-queue-graphs-bin'} = '/usr/bin/fci-generate-queue-graphs.pl';

   $args{'site-data-dir'}		= $args{'settings-dir'}.'site-data/';
   $args{'site-config-file'}		= $args{'site-data-dir'}.'config.xml';
   $args{'site-defaults-file'}		= $args{'site-data-dir'}.'defaults.xml';
   $args{'site-permissions-file'}	= $args{'site-data-dir'}.'permissions.xml';
   
   $args{'useragent'}                   = 'fci-server/'.$args{'version'};

   $args{'demo'}			= 0;
   
   $args{'target'} 			= 'home.plc' 	if(!$args{'target'});
   $args{'action'} 			= '' 		if(!$args{'action'});
   $args{'show'} 			= '' 		if(!$args{'show'});
   $args{'show-inactive-usernames'}	= 0             if(!$args{'show-inactive-usernames'});
   $args{'show-inactive-teams'}		= 0             if(!$args{'show-inactive-teams'});
   $args{'show-inactive-clients'}	= 0             if(!$args{'show-inactive-clients'});
   $args{'show-inactive-projects'}	= 0             if(!$args{'show-inactive-projects'});

my $config = '';
my $error  = '';
if(-r $args{'site-config-file'}) {
	# load settings from config.xml
	$config = read_xml( %args,
	                    Name => 'site-config'
	                  ) || die "Failed to read site-config XML file!\n";

	#print Dumper($config);
	
	# update application paths if valid
	foreach(keys %{$config->{settings}->{application_list}->{application}}) {
		if(valid_application($_, $config, %args)) {
			%args = update_application($_, $config, %args);
		}
		else {
			$error .= "Error: Invalid Application! ($_)\n";
		}
	}
}

# Only check authentication if action == upload and this index.pl is not 
# in the settings/ directory
if( $cwd !~ /^(.*\/)settings\/$/ && 
    $args{target} eq 'home.plc' &&
    $args{action} eq 'upload' &&
    $args{Method} eq 'POST' &&
    $config->{settings}->{authentication}->{site}->{main}->{enabled} eq 'true' &&
    $config->{settings}->{authentication}->{site}->{main}->{scope} eq 'upload'
  ) {
  	if(!$args{'fci-version'} && $args{'client_version'}) {	
		# dealing with pre-1.0 client
		
		if(!valid_authentication(%args, 'pre-1.0 client' => 1)) {
			# supplied username & password are incorrect

			return send_header(401, $r);
		}
	}
	elsif($args{'fci-version'} && substr($args{'fci-version'}, 0, 3) >= 1.0) {
		# clients starting from fci-1.0/fci-client v1.0
		# fci-client.pl v1.0 can deal with 401 responses to its POST request
	
		# Authentication required
		if(!$args{Authorization}) {
			# Client has not supplied username+password, ask for them

			return send_header(401, $r);
		}
		else {
			# Client has send username+password, verify in .htpasswd
			
			if(!valid_authentication(%args)) {
				# supplied username & password are incorrect
		
				return send_header(401, $r);
			}
		}
	}
	else {
		#if($args{'client_version'} && substr($args{'client_version'}, 0, 3) < 0.5) {}
		# clients prior to fci-0.7/fci-client v0.5 are not supported
		# OR
		# really old fci-client.pl version or something else unexpected
	
		return send_header(401, $r);
	}
}

# send http header and page content
$r->header_out('Pragma'        => 'no-cache');
$r->header_out('Cache-control' => 'no-cache');
$r->header_out('Connection'    => 'close');

# X-FCI-Version
$r->header_out('X-FCI'         => "To boldly fold that which has not done so before!");

$r->content_type('text/html');
$r->send_http_header;

# If the UserAgent only requested the HTTP header (using HEAD perhaps),
# we should send it and then exit, skipping the rest of the HTML document.
return 'OK' if $r->header_only;

if($error ne '') {
	print "<pre>";
	print $error;
	print "Error(s) encountered, stopping here!\n";
	print "</pre>\n";
	exit 1;
}

#foreach(keys %args) { print "$_ : $args{$_}\n"; } exit;

# check the permissions of the files and directries, display the errors if any
if(!call('permissions.plc', %args)) {
	print "Permission error(s) encountered, stopping here!\n";
	exit 1;
}

# invoke the upload mode if the action parameter is set to upload
# invoke the display mode otherwise
if($args{action} eq 'upload' && $args{Method} eq 'POST') {	
	call('upload.plc', %args);
}
else {
	call('site.plc', %args);
}

my $time_stop = new Benchmark;
my $time_diff = timediff($time_stop, $time_start);

if($args{action} eq 'upload' && $args{Method} eq 'POST') {
	print "Execution time: ". timestr($time_diff) ."\n";
}
else {
	print "<!-- Execution time: ". timestr($time_diff) ." //-->\n";
}


################################[ Subroutines ]################################

##############
#
#     Routine: dump_var()
#  Parameters: variable
#     Returns: nothing
# Description: prints a variable using Dumper inclosed in <pre> tags
#

sub dump_var {
	my $var = shift;

	print "<pre>";
	print Dumper($var);
	print "</pre>";
}

##############
#
#     Routine: get_option()
#  Parameters: %args
#     Returns: %option
# Description: create %option for construct_url(), preserving global parameters
#

sub get_option {
	my %args = @_;

	my %option = ( 'param' => {} );

        if($args{'show-inactive-usernames'} == 1) {
                $option{'param'}{'show-inactive-usernames'} = 1;
        }
        if($args{'show-inactive-teams'} == 1) {
                $option{'param'}{'show-inactive-teams'} = 1;
        }
        if($args{'show-inactive-clients'} == 1) {
                $option{'param'}{'show-inactive-clients'} = 1;
        }
        if($args{'show-inactive-projects'} == 1) {
                $option{'param'}{'show-inactive-projects'} = 1;
        }

	if($args{anchor}) {
		$option{'anchor'} = $args{anchor};
	}

	return %option;
}

##############
#
#     Routine: construct_url()
#  Parameters: path
#              target
#              options hashref
#              parameters hash
#     Returns: url string
# Description: create in in-application URL preserving global parameters
#

sub construct_url {
	my $path   = shift;
	my $target = shift;
	my $option = shift;
	my %param  = @_;

	$path   = '' if(!$path);
	$target = '' if(!$target);

	# path is the default URL
	my $url = $path;

	my $i = 0;
	# taget is expected, but not required
	if($target && $target =~ /^[\w\-]+\.plc$/) {
		$url .= $i == 0 ? '?' : '&amp;';
		$url .= "target=". uri_escape($target);
		$i++;
	}

	# required parameters
	foreach(sort keys %param) {
		$url .= $i == 0 ? '?' : '&amp;';
		$url .= uri_escape($_) .'='. uri_escape($param{$_});
		$i++;
	}

	# optional parameters
	foreach(sort keys %{$option->{param}}) {
		$url .= $i == 0 ? '?' : '&amp;';
		$url .= uri_escape($_) .'='. uri_escape($option->{param}->{$_});
		$i++;
	}

	# optional anchor
	if($option->{anchor}) {
		$url .= '#'.$option->{anchor};
	}

	return $url;
}

##############
#
#     Routine: menu_item()
#  Parameters: target
#              name
#              title
#              %args
#     Returns: nothing
# Description: print a menu item link
#

sub menu_item {
        my ($target, $name, $title, %args) = @_;

        my $indent = '';
           $indent = $args{indent} if($args{indent});

	my $i = 0;
        my %param  = ();
	my %option = ();
        foreach(keys %args) {
                if(/^param_(.*)$/) {
                        $param{$1} = $args{$_};
			$i++;
                }
		elsif(/^option_param_(.*?)$/) {
			$option{'param'}{$1} = $args{$_};
		}
		elsif(/^option_(.*?)$/) {
			$option{$1} = $args{$_};
		}
        }

        my $selected = 2;
           $selected = 0 if($args{unselectable});

	my $url = construct_url('', $target, \%option, %param);
	
	if($i > 0) {
		# menu target with required parameters 
 	       	foreach(keys %param) {
	                $args{$_} = '' if(!$args{$_});
	                if($args{$_} eq $param{$_}) {
	                        $selected = 1 unless($selected == 0);
	                }
	                else {
	                        $selected = 0;
	                }
		}
	}
	else {
		# menu target without required parameters
		if($args{target} eq $target) {
			$selected = 1 unless($selected == 0);
		}
	}

        print "\t$indent<a href='$url'";
        if($args{target} eq $target && $selected == 1) {
                print " class='selected'";
        }
        print " title='$title'" if($title);
        print ">$name</a><br>\n";
}

##############
#
#     Routine: username_link()
#  Parameters: username name
#  	       username hash
#  	       optional parameters hash
#     Returns: 1 on succes, 0 on failure
# Description: print a link to a username page and optional state markings
#

sub username_link {
	my $name     = shift;
	my $username = shift;
	my $option   = shift;

	if(!defined $name) {
		return 0;
	}
	elsif(!($username->{team} >= 0)) {
		return 0;
	}

	my $url = construct_url('', 'usernames.plc', $option, 'username' => $name, 
							      'team'     => $username->{team});

	print "<a href='$url' title='Detailed username information'>$name</a>";

	if(exists $username->{state} && $username->{state} eq 'inactive') {
		print "<span title='Inactive username (24 hour average is zero)'";
		print " class='inactive'>&nbsp;I&nbsp;</span>";
	}

	return 1;
}

##############
#
#     Routine: team_link()
#  Parameters: team number
#              team name
#  	       team hash
#  	       optional parameters hash
#     Returns: 1 on succes, 0 on failure
# Description: print a link to a team page and optional state markings
#

sub team_link {
	my $number = shift;
	my $name   = shift;
	my $team   = shift;
	my $option = shift;

	if(!defined $number) {
		return 0;
	}
	if(!defined $name || $name eq '') {
		$name = $number;
	}

	my $url = construct_url('', 'teams.plc', $option, 'team' => $number);

	print "<a href='$url' title='Detailed team information'>$name</a>";

	if(exists $team->{state} && $team->{state} eq 'inactive') {
		print "<span title='Inactive team (24 hour average is zero)'";
		print " class='inactive'>&nbsp;I&nbsp;</span>";
	}
	else {
		print "<span title='Active team (24 hour average is $team->{points_24hr_avg})'";
		print " class='invisible'>&nbsp;I&nbsp;</span>";
	}

	return 1;
}

##############
#
#     Routine: client_link()
#  Parameters: client name
#  	       client hash
#  	       optional parameters hash
#     Returns: 1 on succes, 0 on failure
# Description: print a link to a client page and optional state markings
#

sub client_link {
	my $name   = shift;
	my $client = shift;
	my $option = shift;

	if(!defined $name) {
		return 0;
	}

	my $url = construct_url('', 'clients.plc', $option, 'client' => $name);

	print "<a href='$url' title='Detailed client information'>$name</a>";

	if(exists $client->{state} && $client->{state} eq 'inactive') {
		print "<span title='Inactive client (no upload sice ";
		print $client->{last_update};
		print ")' class='inactive'>&nbsp;I&nbsp;</span>";
	}
	elsif(exists $client->{state} && $client->{state} eq 'new') {
		print "<span title='New client (first upload on ";
		print $client->{last_update};
		print ")' class='new'>&nbsp;*&nbsp;</span>";
	}

	if(exists $client->{current_work_unit}->{progress_state} &&
	   $client->{current_work_unit}->{progress_state} eq 'critical'
	) {
		print "<span title='Progress state critical (expected after final deadline)'";
		print " class='critical'>&nbsp;!&nbsp;</span>";
	}
	elsif(exists $client->{current_work_unit}->{progress_state} &&
	      $client->{current_work_unit}->{progress_state} eq 'warning'
	) {
		print "<span title='Progress state warning (expected after preferred deadline)'";
		print " class='warning'>&nbsp;!&nbsp;</span>";
	}
	else {
		print "&nbsp;";
	}

	return 1;
}

##############
#
#     Routine: project_link()
#  Parameters: project name
#  	       project hash
#  	       optional parameters hash
#     Returns: 1 on succes, 0 on failure
# Description: print a link to a project page and optional state markings
#

sub project_link {
	my $number  = shift;
	my $project = shift;
	my $option  = shift;

	if(!defined $number) {
		return 0;
	}

	my $url = construct_url('', 'projects.plc', $option, 'project' => $number);

	print "<a href='$url' title='Detailed project information'>$number</a>";

	if(exists $project->{state} && $project->{state} eq 'inactive') {
		print "<span title='Inactive project (not listed on the \"Currently Running Projects\" page";
		print " &amp; no clients working on it)' class='inactive'>&nbsp;I&nbsp;</span>";
	}
	else {
		print "<span title='Active project (listed on the \"Currently Running Projects\" page";
		print " or has clients working on it)' class='invisible'>&nbsp;I&nbsp;</span>";
	}

	return 1;
}

##############
#
#     Routine: update_application()
#  Parameters: application name
#              configuration file XML structure
#              %args
#     Returns: 1 on success or 0 on error
# Description: validate an application path
#

sub update_application {
	my $application = shift;
	my $config      = shift;
	my %args        = @_;

	if(exists $config->{settings}->{application_list}->{application}->{$application} && 
	   exists $config->{settings}->{application_list}->{application}->{$application}->{path}
	) {
		#print "Update: $_ -> $config->{settings}->{application_list}->{application}->{$application}->{path}\n";
		$args{$_.'-bin'} = $config->{settings}->{application_list}->{application}->{$application}->{path};
	}

	return %args;
}

##############
#
#     Routine: valid_application()
#  Parameters: application name
#              configuration file XML structure
#              %args
#     Returns: 1 on success or 0 on error
# Description: validate an application path
#

sub valid_application {
	my $application = shift;
	my $config      = shift;
	my %args        = @_;

	# no entry in the XML file
	if(!exists $config->{settings}->{application_list}->{application}->{$application}) {
		#print "Error: No entry in XML file for: $application\n";
		return 0;
	}
	if(!$config->{settings}->{application_list}->{application}->{$application}) {
		#print "Error: Application cannot be empty!\n";
		return 0;
	}
	if(!exists $config->{settings}->{application_list}->{application}->{$application}->{path}) {
		#print "Error: No entry in XML file to path!\n";
		return 0;
	}
	if(!$config->{settings}->{application_list}->{application}->{$application}->{path}) {
		#print "Error: Path cannot be empty!\n";
		return 0;
	}
	if(!-x $config->{settings}->{application_list}->{application}->{$application}->{path}) {
		#print "Error: Cannot execute: $application\n";
		#print $config->{settings}->{application_list}->{application}->{$application}->{path}."\n";
		return 0;
	}

	return 1;
}


##############
#
#     Routine: read_xml()
#  Parameters: %args
#     Returns: a hashreference of the XML structure
# Description: read an XML file while enforcing the correct structure
#

sub read_xml {
	my %args = @_;
	
	if(!$args{Name}) {
		warn "FCI read_xml: Unknown name!\n";	
		return undef;
	}

	if($args{Name} =~ /^(client|qd|team-stats)$/) {
		if(!$args{File}) {
			warn "FCI read_xml: No file specified! ($args{Name})\n";
			return undef;
		}
	}
	else {
		if(!$args{$args{Name}.'-file'} ) {
			warn "FCI read_xml: Unknown file: $args{Name}\n";
			return undef;	
		}
		if(!$args{File}) {
			$args{File} = $args{$args{Name}.'-file'};
		}
	}
	
	if(!-e $args{File}) {
		warn "FCI read_xml: File does not exist: $args{File}\n";
		return undef;
	}
	if(-e $args{File} && !-r $args{File}) {
		warn "FCI read_xml: Cannot read file: $args{File}\n";
		return undef;
	}
	if(!-f $args{File}) {
		warn "FCI read_xml: Not a file: $args{File}\n";
		return undef;
	}
	
	if($args{File} && -r $args{File} && -f $args{File}) {
		if($args{Name} eq 'qd-files') {
			return XMLin( $args{File},
			              KeyAttr    => { file => 'id' },
				      ForceArray => [ 'file' ],
				      KeepRoot   => 1,
				    ) || die $@;
		}
		elsif($args{Name} eq 'team-stats') {
			return XMLin( $args{File},
			              KeyAttr    => 'team_rank',
				      ForceArray => [ 'rank_entry' ],
				      KeepRoot   => 1,
				    ) || die $@;
		}
		elsif($args{Name} eq 'teamstats-list') {
			return XMLin( $args{File},
			              KeyAttr    => { team_stats => 'team' }, 
				      ForceArray => [ 'team_stats' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'stanford-files') {
			return XMLin( $args{File},
			              KeyAttr    => { file => 'id' }, 
				      ForceArray => [ 'file' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'client-projects') {
			return XMLin( $args{File},
			              KeyAttr    => { project  => 'number', client => 'name' }, 
				      ForceArray => [ 'project', 'client' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'project-images') {
			return XMLin( $args{File},
			              KeyAttr    => { image  => 'type',  setting  => 'id', 
				                      option => 'value', location => 'id' },
				      ForceArray => [ 'image', 'setting', 'option', 'location' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'known-projects') {
			return XMLin( $args{File},
				      KeyAttr    => { project => 'number', image => 'type' },
				      ForceArray => [ 'project', 'image' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'site-defaults') {
			return XMLin( $args{File},
			              KeyAttr    => { application => 'name', site => 'name' },
				      ForceArray => [ 'application', 'site' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'site-config') {
			return XMLin( $args{File},
			              KeyAttr    => { application => 'name', site => 'name' },
				      ForceArray => [ 'application', 'site' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'site-permissions') {
			return XMLin( $args{File},
			              KeyAttr    => { path => 'name' },
				      ForceArray => [ 'path' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'qd') {
			return XMLin( $args{File},
				      KeyAttr    => { index => 'number', server => 'number' },
				      ForceArray => [ 'index', 'server' ]
				    ) || die $@;
		}
		elsif($args{Name} eq 'client') {
			return XMLin( $args{File},
				      KeyAttr    => { client => 'name', 'server' => 'number' },
				      ForceArray => [ 'server' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'client-list') {
			return XMLin( $args{File},
				      KeyAttr    => { client => 'name' },
				      ForceArray => [ 'client' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'username-list') {
			return XMLin( $args{File},
			              KeyAttr    => { username => 'userid' },
				      ForceArray => [ 'username' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'team-list') {
			return XMLin( $args{File},
			              KeyAttr    => { team => 'number', username => 'userid' },
				      ForceArray => [ 'team', 'username' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'xml-project-summary') {
			return XMLin( $args{File},
			              KeyAttr    => { project => 'number' },
				      ForceArray => [ 'project' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'fci-summary') {
			return XMLin( $args{File},
				      KeepRoot   => 1
				    ) || die $@;
		}
                elsif($args{Name} eq 'jmol-missing-projects') {
                        return XMLin( $args{File},
                                      ForceArray => [ 'project' ],
                                      KeepRoot   => 1
                                    ) || die $@;
                }
		elsif($args{Name} eq 'jmol-projects') {
			return XMLin( $args{File},
				      KeyAttr    => { fah_proj => 'number' },
				      ForceArray => [ 'fah_proj' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
                elsif($args{Name} eq 'queue-graphs') {
                        return XMLin( $args{File},
                                      ForceArray => [ 'client' ],
                                      KeepRoot   => 1
                                    ) || die $@;
                }
		else {
			# unknown file
			return undef;
		}
	}

	return undef;
}


##############
#
#     Routine: write_xml()
#  Parameters: name of the XML file 
#              xml structure in a hash reference
#              %args
#     Returns: a hashreference of the XML structure
#     Returns: 1 on sucess
#              0 and string with reasons on errors
# Description: write an XML file while enforcing the correct structure
#

sub write_xml {
	my %args = @_;
	
	my ($success, $reason) = (1, '');
	
	if(!$args{Name}) {
		return (0, "Unknown name!\n");
	}
	
	if($args{Name} =~ /^(client|qd)$/) {
		if(!$args{File}) {
			return (0, "No file specified! ($args{Name})\n");
		}
	}
	else {
		if(!$args{$args{Name}.'-file'} ) {
			return (0, "Unknown file: $args{Name}\n");
		}
		if(!$args{File}) {
			$args{File} = $args{$args{Name}.'-file'};
		}
	}
	
	if(!-e $args{File} && !-w dirname($args{File}) ) {
		return (0, "Cannot write directory for: $args{Name}\n");
	}
	if(-e $args{File} && !-w $args{File}) {
		return (0, "Cannot write file: $args{Name}\n");
	}
	if(-e $args{File} && !-f $args{File}) {
		return (0, "Not a file: ($args{Name})\n");
	}
	
	
	if($success) {
		if($args{Name} eq 'qd-files') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { file => 'id' },
				NoAttr     => 1,
				KeepRoot   => 1,
		              ) || die $@;
		}
		elsif($args{Name} eq 'team-stats') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => 'team_rank',
				NoAttr     => 1,
				KeepRoot   => 1,
		              ) || die $@;
		}
		elsif($args{Name} eq 'teamstats-list') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { team_stats => 'team' }, 
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'stanford-files') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { file => 'id' }, 
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'client-projects') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
			        KeyAttr    => { project  => 'number', client => 'name' }, 
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'project-images') {
			XMLout( $args{XML},
				OutputFile => $args{File},
				KeyAttr    => { image  => 'type',  setting  => 'id', 
				                option => 'value', location => 'id' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'known-projects') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
			        KeyAttr    => { project => 'number', image => 'type' },
				NoAttr     => 1,
				KeepRoot   => 1
		              ) || die $@;
		}
		elsif($args{Name} eq 'site-defaults') {
			XMLout( $args{XML},
				OutputFile => $args{File},
				KeyAttr    => { application => 'name', site => 'name' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'site-config') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { application => 'name', site => 'name' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'site-permissions') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
			        KeyAttr    => { path => 'name' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'client') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { client => 'name', server => 'number', 'fci' => 'version' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'client-list') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { client => 'name' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'username-list') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { username => 'userid' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'team-list') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { team => 'number', username => 'userid' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'xml-project-summary') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { project => 'number' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'jmol-missing-projects') {
			XMLout( $args{XML},
				OutputFile => $args{File},
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'jmol-projects') {
			XMLout( $args{XML},
				OutputFile => $args{File},
				KeyAttr    => { fah_proj => 'number' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
                elsif($args{Name} eq 'queue-graphs') {
			XMLout( $args{XML},
				OutputFile => $args{File},
				KeyAttr    => { client => 'name' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		else {
			# unknown file
			$success = 0;
			$reason .= "Unknown file: $args{Name}\n";
		}
	}

	return ($success, $reason);
}


##############
#
#     Routine: normalize_filesize()
#  Parameters: filesize
#              current size unit
#     Returns: normalized filesize and unit name
# Description: convert a filesize to its smallest unit
#     Example: 2097152 KB -> 2 GB
#

sub normalize_filesize {
	my $size = $_[0] =~ /^\d+$/        ? $_[0] : 0;
	my $unit = $_[1] =~ /^[KMGTPE]?B$/ ? $_[1] : 'B';

	my $u     = 0;
	my @units = qw/B KB MB GB TB PB EB/;
	foreach(@units) {
		if($unit =~ /^$_/i) {
			if($size >= 1024) {
				$size /= 1024;
				$unit  = $units[$u+1];
			}
			else {
				last;
			}
		}
		$u++;
	}

	return ($size, $unit);
}


##############
#
#     Routine: valid_authentication()
#  Parameters: .htpasswd path
#              Authorization HTTP header or Username & Password
#     Returns: 1 on valid username+password combination
#              0 otherwise
# Description: Check the authentication of a client, either by the 
#              Authorization HTTP header or the username & password supplied in 
#              the form data

sub valid_authentication {
	my %args = @_;
	
	my $htpasswd = new Apache::Htpasswd( { passwdFile => $args{'htpasswd-file'},
        	       	                       ReadOnly   => 1 } );
  	
	my $username = '';
	my $password = '';
	
	if($args{'pre-1.0 client'}) {
		# clients before fci-client.pl v1.0 sent the username+password as form fields
		
		$username = $args{username};
		$password = $args{password};
	}
	else {
		# fci-client.pl v1.0 and beyond uses proper HTTP Authentication
	
		# Basic H6Hu8n87==
		my $auth = $args{Authorization};
		if($auth =~ s/Basic (.*?)$/$1/g) {}
		   $auth = decode_base64($auth);
	
		($username, $password) = split /:/, $auth;
	}

	if($htpasswd->htCheckPassword($username, $password)) {
		return 1;
	}

	return 0;
}
	

sub iso_date {
	my $date     = $_[0];
	my $iso_date = $date;
	
	# Tue Jan 17 13:37:47 2006
	if($date =~ /^(\w+)\s+(\w+)\s+(\d+)\s+(\d+:\d+:\d+)\s+(\d+)$/) {
		my $dow   = $1;
		my $month = $2;
		my $day   = $3;
		my $time  = $4;
		my $year  = $5;
		
		$month = Date::Calc::Decode_Month($month);
		
		$day   = '0'.$day   if(length($day)   == 1);
		$month = '0'.$month if(length($month) == 1);
		
		$iso_date = "$year-$month-$day $time";
	}
	
	return $iso_date;
}


##############
#
#     Routine: time_format()
#  Parameters: time in seconds
#     Returns: time in semi human readable format: 1d2h3m4s
# Description: convert time in seconds to readable format
#

sub time_format {
        my $value = shift;
        my $ret;

        if($value && ($value >= 0)) {
                # convert time in seconds to human readable format
                # e.g. 65 seconds == 1m5s

                my $date1 = ParseDate("today");
                my $date2 = DateCalc($date1, "+ $value seconds");

                my $delta = DateCalc($date1, $date2);
                #print "Delta: $delta\n";

                #+0:0:0:0:1:55:43
                if($delta =~ /(\d+):(\d+):(\d+):(\d+):(\d+):(\d+):(\d+)$/) {
                        #+Y:+M:+W:+D:+H:+MN:+S
                        my $years   = $1;
                        my $months  = $2;
                        my $weeks   = $3;
                        my $days    = $4;
                        my $hours   = $5;
                        my $minutes = $6;
                        my $seconds = $7;

                        my $time  = '';
			   $time .= $years  ."y"   if($years   > 0);
                           $time .= $months ."m"   if($months  > 0);
                           $time .= $days   ."d"   if($days    > 0);
                           $time .= $hours  ."h"   if($hours   > 0);
                           $time .= $minutes."m"   if($minutes > 0);
                           $time .= $seconds."s"   if($seconds > 0);
                        
			$ret = $time;
                }
                else {
                        $ret = $value;
                }
        }
        else {
		$ret = $value;
        }

        return $ret;
}


##############
#
#     Routine: read_file()
#  Parameters: path to file
#     Returns: content of the file
# 	       undef in case of error
# Description: return the contents of a file
#

sub read_file {
	my $file    = $_[0];
	my $verbose = $_[1] ? $_[1] : 0;
	
	print "Reading file from disc: $file ... " if($verbose);
	if(-r $file) {
		my $data = '';

		open(F, "< $file") || return undef;
		while(<F>) {
			$data .= $_;
		}
		close F;

		print "Done.\n" if($verbose);

		return $data;
	}
	else {
		print "FAILED!\n" if($verbose);
		return undef;
	}
}


##############
#
#     Routine: save_file()
#  Parameters: path to file
# 	       data to save into the file
# 	       optional hash with
# 	         mode    => file premission in octal
# 	         binary  => flag to set binary mode (only needed on Windows)
# 	         verbose => flag to set verbose output
#     Returns: 1 for successfull save or undef for unsuccessfull save
#              the reason for the failure or an empty string
# Description: write a file to disc if the required permissions are set
#

sub save_file {
	my $filename = shift;
	my $content  = shift;
	my %option   = @_;

	my ($success, $reason) = (1, '');

	print "Saving file to disc: $filename ... " if($option{verbose});
	if(!-e $filename || -w $filename) {
		print "Done.\n" if($option{verbose});
		
		open(F, "> $filename") || return (0, "Error: Cannot save file: $filename ($!)\n"); 
		if($option{binary} && $option{binary} == 1) {
			print "Binmode enabled.\n" if($option{verbose});
			binmode(F);
		}
		print F $content;
		close F;

		if($option{mode} && $option{mode} =~ /^0\d{3}$/) {
			print "Set permission: $option{mode}.\n" if($option{verbose});
			chmod oct($option{mode}), $filename;
		}
		else {
			if($option{verbose}) {
				print "Ignoring invalid mode: ";
				print $option{mode};
				print "\n";
			}
		}

		return 1;
	}
	else {
		print "FAILED! (File is not writeable)\n" if($option{verbose});
		return (0, "Error: File does not exists or is unwriteable!\n");
	}
}


################[ Start helper subroutines for call() routine ]################

##############
#
#     Routine: call()
#  Parameters: path to .plc file
#              global hash
#     Returns: return value of call .plc file (script)
#              undef in case of error
# Description: evaluate a perl script component, the perl code is embedded between 
#              regular html code
#

sub call {
	# for security reseason, only allow scripts to be called in the current 
	# directory, the one which holds this index.pl script
	#
	# parse the script, show errors if any, and error messages are allowed 
	# to be displayed (debug/warning mode)
	#
	# eval() the parsed Perl code, show errors if any and warning mode is on
	#
	# execute the eval()'ed code, show errors if any and are allowed to be shown
	#
	# return the return value if any
	#
	# TODO: implement switch to disable printing of error messages

	my ($file, %args) = @_;
	
	if($file =~ /\.plc$/ && $file !~ /\// && -r "./$file") {
		if(my $script = parse($file)) {
			my $code = eval $script;
			if($@ || !$code) {
				print "<div style='color: #FF0000'>Error compiling $file <pre>$@</pre></div><br><hr>\n";
				return undef;
			}
			else {
				my $return;
				eval {
					$return = &$code(%args);
				};
				
				if($@) {
					print "<div style='color: #FF0000'>Error running $file: <pre>$@</pre></div><br><hr>\n";
				}
				else {
					return $return;
				}
			}
		}
		else {
			print "<div style='color: #FF0000'>Cannot load $file</div><br><hr>\n";
			return undef;
		}
	}
	else {
		print "<div style='color: #FF0000'>Cannot load $file !!!</div><br>\n";
		return undef;
	}
}


##############
#
#     Routine: parse()
#  Parameters: path to .plc file
#     Returns: sub routine of compiled perl code from the .plc file
#              undef in case of error
# Description: evaluate a perl script component, the perl code is embedded between 
#              regular html code

sub parse {
	# compile all the Perl code sections from the requested file into a 
	# single piece of code
	# 
	# return the code as a subroutine to be executed
	# return undef if no code is compiled
	
	my $file  = $_[0];
	my $parse = 0;
	my $data  = '';

	open(F, $file) || return undef;
	while(<F>) {
		if(/<\%perl>/i && !$parse) {
			$parse = 1;
			$data .= "\n";
		}
		elsif(/<\/\%perl>/i && $parse) {
			$parse = 0;
			$data .= "\n";
		}
		elsif(/<\%/ && !$parse) {
			while(/<%(.*?)%>/) {
				my $code    = $1;
				my $prefix  = $`;
				my $postfix = $';
				
				$prefix =~ s/'/'."'".'/g;
				$data  .= "print '$prefix';";
				$data  .= "print $code;";
				
				$_ = $postfix;
			}
			s/'/'."'".'/g;
			$data .= "print '$_';";
		}
		else {
			if($parse) {
				$data .= $_;
			}
			else {
				s/'/'."'".'/g;
				$data .= "print '$_';";
			}
		}
	}
	close F;
	
	if($data) {
		return 'return sub { my %args = @_; '.$data.' };';
	}
	else {
		return undef;
	}
}

#################[ End helper subroutines for call() routine ]#################

sub send_header {
	my $code = shift;
	my $r    = shift;
	
	if($code == 401) {
		$r->err_headers_out->add('WWW-Authenticate' => 'Basic realm="Folding@Home Client Info"');
		$r->status(401);
		
		$r->content_type('text/plain');
		$r->send_http_header;

		return 'AUTH_REQUIRED';
	}
}


##############
#
#     Routine: get_date()
#  Parameters: optionally a hash with the keys date and/or time set to something
#     Returns: the date in yyyy-mm-dd format, if only date is set
#              the time in hh:mm:ss format, if only time is set
#              the full date in yyyy-mm-dd hh:mm:ss format, if neither date and 
#              time are set
#              both the date and the time, if they're both set
# Description: Get the contents and other info of uploaded data
#   

sub get_date {
	my %option = @_;

	my ($year,$month,$day,$hour,$min,$sec) = Date::Calc::Today_and_Now();
	
	$month = '0'.$month if(length($month) == 1);
	$day   = '0'.$day   if(length($day)   == 1);
	$hour  = '0'.$hour  if(length($hour)  == 1);
	$min   = '0'.$min   if(length($min)   == 1);
	$sec   = '0'.$sec   if(length($sec)   == 1);
	
	my $date = "$year-$month-$day";
	my $time = "$hour:$min:$sec";

	if(!$option{date} && !$option{time}) {
		return "$date $time";
	}
	elsif($option{time} && !$option{date}) {
		return $time;
	}
	elsif($option{date} && !$option{time}) {
		return $date;
	}
	else {
		return ($date, $time);
	}
}

##############
#
#     Routine: upload_properties()
#  Parameters: Apache::Upload object
#              Parameter name
#              %args
#     Returns: updated hash with HTTP headers, URL variables, and other global variables
# Description: Get the contents and other info of uploaded data
#   

sub upload_properties {
        my $upload = shift;  # upload object
	my $param  = shift;  # parameter name
	my %args   = @_;
	
	# Uploaded file properties
	$args{$param.'_Size'}     = $upload->size;
	$args{$param.'_Type'}     = $upload->type;
	$args{$param.'_Tempname'} = $upload->tempname;
	$args{$param.'_Filename'} = $upload->filename;

	# Save file content
	my $fh = $upload->fh;
	while (<$fh>) {
		$args{$param} .= $_;
	}

	# strip the filename from the start of the data
	if($args{$param} =~ s/^\Q$args{$param.'_Filename'}\E//) {}

	# handle files uploaded with Internet Explorer
	# instead of just the filename, IE sends the whole path
	# example:
	# X:\Porn\Pics\NakedLady.jpg
	
	if($args{$param.'_Filename'} =~ /^\w:.*\\(.*?)$/) {
		$args{$param.'_Filename'} = $1;
	}
	
	return \%args;
}
						
	
##############
#
#     Routine: get_args()
#  Parameters: Apache::Request object
#              Apache::Connection object
#              Apache::Server object
#              Apache2 flag
#     Returns: hash with HTTP headers, URL variables, and other global variables
# Description: get the information from the HTTP request and Apache API, 
#              and set global variables
#   
   
sub get_args {
	my ($r, $c, $s, $apache2) = @_;
	my %args;

	# fill hash with URI parameters

	my $apr;
	if($apache2 != 1) {
		$apr = Apache::Request->new($r);
	}
	else {
		$apr = Apache2::Request->new($r);
	}

	# normal parameters
	foreach my $param ($apr->param) {
		my $i = 0;
		foreach my $value ($apr->param($param)) {
			# parameters can have multiple values,
			# number the ones afther the 1st
			
			$param .= $i if($i > 0);
			
			$args{$param} = $value;
			
			$i++;
		}
	}

	# handle POST requests
	if($r->method eq 'POST') {
		# uploaded files
		if($apache2 != 1) {
			for my $upload ($apr->upload) {
				my $param = $upload->name;
				
				foreach(keys %{$upload->info}) {
					$args{$param.'_Info_'.$_} = $upload->info($_);
				}
				
				%args = %{upload_properties($upload, $param, %args)};
			}
		}
		else {
			my $upload_table_ref = $apr->upload; # APR::Table ref of uploads
			
			foreach my $name (keys %$upload_table_ref) {
				my $i = 0;
				foreach my $upload ($apr->upload($name)) {
					my $param = $upload->name;
					
					# parameters can have multiple values,
					# number the ones afther the 1st

					$param .= $i if($i > 0);

					
					foreach(keys %{$upload->info}) {
						$args{$param.'_Info_'.$_} = $upload->info->{$_};
					}
					
					%args = %{upload_properties($upload, $param, %args)};
					
					$i++;
				}
			}
		}
	}
	
	my ($date, $time) = get_date(date => 1, time => 1);

	$args{'Date'}		= $date;
	$args{'Time'}		= $time;
	$args{'User'}           = $c->user		if defined($c->user);
	$args{'Protocol'}   	= $r->protocol;
	
	$args{'Bytes'}		= $r->bytes_sent	if($r->bytes_sent);
	$args{'Method'}		= $r->method;
	$args{'MethodNumber'}   = $r->method_number;
	$args{'Request'}        = $r->the_request;
	$args{'RequestTime'}    = $r->request_time;
	$args{'URI'}            = $r->uri;
	$args{'Filename'}       = $r->filename;
	$args{'PathInfo'}       = $r->path_info		if($r->path_info);
	$args{'DocumentRoot'}   = $r->document_root;
	
	$args{'UserAgent'}      = $r->header_in('User-Agent');	# for compatibility with old scripts	

	foreach(keys %{$r->headers_in()}) {
		$args{$_} = $r->headers_in->{$_};
	}
	
	$args{'Hostname'}   	= $r->hostname;
	$args{'AuthName'}	= $r->auth_name		if($r->auth_name);
	$args{'AuthType'}	= $c->auth_type		if($c->auth_type);
	$args{'RemoteIP'}	= $c->remote_ip;
	if($c->remote_host() eq '') {
		$args{'RemoteHostname'} = $r->get_remote_host;
	}
	else {
		$args{'RemoteHostname'} = $c->remote_host;
	}

	$args{'ServerName'}     = $s->server_hostname;
	$args{'ServerAdmin'}    = $s->server_admin;
	$args{'ServerPort'}     = $s->port;
	$args{'ServerSoftware'} = $ENV{SERVER_SOFTWARE};	

	return %args;
}


##############
#
#     Routine: html_footer()
#  Parameters: hash of tags to generate
#     Returns: nothing
# Description: print the tags of the header of the html page
#   

sub htmlheader {
	my %args = @_;
	
	if($args{doctype}) {
		print "<!DOCTYPE $args{doctype} PUBLIC \"-//W3C//DTD ";
		if($args{doctype_dtd}) {
			print "$args{doctype_dtd}";
		}
		else {
			print "$args{doctype}";
		}
		
		if($args{doctype_language}) {
			print "//$args{doctype_language}\"";
		}
		else {
			print "//EN\"";
		}

		if($args{doctype_dtd_url}) {
			print " \"$args{doctype_dtd_url}\"";
		}
		print ">\n";
	}
	
	print "<html>\n";
	print "<head>\n";
	print "\t<title>$args{title}</title>\n" if($args{title});
	if($args{meta1_name}) {
		my $i = 1;
		while($args{'meta' . $i . '_name'}) {
			print "\t<meta";
			
			my $meta_name = $args{'meta' . $i . '_name'} if($args{'meta' . $i . '_name'});
			print " name=\"$meta_name\"" if($args{'meta' . $i . '_name'});
			
			my $meta_equiv = $args{'meta' . $i . '_equiv'} if($args{'meta' . $i . '_equiv'});
			print " http-equiv=\"$meta_equiv\"" if($args{'meta' . $i . '_httpequiv'});
			
			my $content  = $args{'meta' . $i . '_content'} if($args{'meta' . $i . '_content'});
			print " content=\"$content\"" if($args{'meta' . $i . '_content'});

			print ">\n";
			
			$i++;
		}
	}
	if($args{link1}) {
		my $i = 1;
		while($args{'link' . $i}) {
			print "\t<link";
			
			my $linktype = $args{'link' . $i . '_type'} if($args{'link' . $i . '_type'});
			print " type=\"$linktype\"" if($args{'link' . $i . '_type'});
			
			my $linkrel  = $args{'link' . $i . '_rel'} if($args{'link' . $i . '_rel'});
			print " rel=\"$linkrel\"" if($args{'link' . $i . '_rel'});
			
			print " href=\"$args{'link' . $i}\">\n";

			$i++;
		}
	}
	if($args{script1}) {
		my $i = 1;
		while($args{'script' . $i}) {
			print "\t<script";
			
			my $scripttype = $args{'script' . $i . '_type'} if($args{'script' . $i . '_type'});
			print " type=\"$scripttype\"" if($args{'script' . $i . '_type'});
			
			my $scriptlanguage  = $args{'script' . $i . '_language'} if($args{'script' . $i . '_language'});
			print " language=\"$scriptlanguage\"" if($args{'script' . $i . '_language'});
			
			print " src=\"$args{'script' . $i}\"></script>\n";
			
			$i++;
		}
	}
	print "\t<style>$args{style}</style>\n" if($args{style});
	
	print "</head>\n\n";
	
	print "<body";
	print " onUnload=\"$args{body_onunload}\"" if($args{body_onunload});
	print " onLoad=\"$args{body_onload}\"" if($args{body_onload});
	print ">\n\n";
}


##############
#
#     Routine: htmlfooter()
#  Parameters: none
#     Returns: nothing
# Description: print the closing tags for the html page
#

sub htmlfooter {
	print "\n</body>\n";
	print "</html>\n";
}
